#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <setjmp.h>
#include "evaldecla.h"

/* Advanced C
 * Chapter 9 Expression Parsing and Evaluation
 *
 * Anders Furuhed adopted it for Prime/Primos
 * Floating point version by Anders Furuhed
 *
 *
 * 871119  Added function support
 * 880916  Removed floating point (not needed in ASM6502...) : Peter Eriksson
 */

/* Simple recursive descent parser for integer expressions

Operators

<, >, <>, =, <=, >=, &, |, ~, +, -, *, /, %, ^, <<, >>, <-, ->,
<- and -> (rotates) are not implemented yet

*/

#define DELIMETER     1
#define VARIABLE      2
#define NUMBER        3
#define FUNCTION      4

#define ROL ( ((x>>15) | (x<<1)) & 65535)
#define ROR ( ((x<<15) | (x>>1)) & 65535)


static char   *evaexpress;    /* Holds expression to be analyzed */
static char   token[80];
static char   tok_type;
static jmp_buf errenv;
static int    (*error)();

char   errmess[100];        /* Global error message variable */
static level4(),
       level5(),
       level6(),
       level7(),
       level8(),
       level9(),
       level10(),
       get_exp(),
       primitive(),
       arith(),
       unary(),
       putback(),
       serror(),
       get_token(),
       iswhite(),
       isdelim(),
       issymbol(),
       is_in(),
       get_int();

static number find_var();

/*
 *  <expr> is the expression to be parsed
 *  The result is placed in <res>
 *  The error messages is sent to <errfunc>. If it's NULL, serror is used
 */
evald(expr, res, errfunc)
  char   *expr;
  number *res;
  int    (*errfunc)();
{
   evaexpress = expr;

   /* Set an error-function */
   if( errfunc == NULL)
     error = &serror;
   else
     error = errfunc;

   if (setjmp(errenv))
      return -1;
   get_exp(res);
   return 0;
}



static get_exp( result)
number *result;
{
  get_token();
  if( !*token)
  {
    (*error)(NO_EXPRE_PRE);
    return -1;
  }
  level4( result);
  return *result;
}

static level4( result)
number *result;
{
  char op[3];
  number hold;

  level5( result);
  while( !strcmp( token, "<") ||
         !strcmp( token, ">") ||
         !strcmp( token, "=") ||
         !strcmp( token, "<=") ||
         !strcmp( token, ">=") ||
         !strcmp( token, "<>"))
  {
    strcpy( op, token);
    get_token();
    level5( &hold);
    arith( op, result, &hold);
  }
}

static level5( result)
  number *result;
{
  char op[3];
  number hold;

  level6( result);
  while( !strcmp( token, "&") ||
         !strcmp( token, "|") ||
         !strcmp( token, "~"))
  {
    strcpy( op, token);
    get_token();
    level6( &hold);
    arith( op, result, &hold);
  }
}

static level6( result)
  number *result;
{
  char op[3];
  number hold;

  level7( result);
  while( *token == '+' || *token == '-')
  {
    strcpy( op, token);
    get_token();
    level7( &hold);
    arith( op, result, &hold);
  }
}

static level7( result)
  number *result;
{
  register char op;
  char ope[3];
  number hold;

  level8( result);
  while( *token == '*' ||
         *token == '%' ||
         *token == '/' ||
         !strcmp( token, "->") ||
         !strcmp( token, "<-") ||
         !strcmp( token, ">>") ||
         !strcmp( token, "<<"))
  {
    strcpy( ope, token);
    get_token();
    level8(&hold);
    arith( ope, result, &hold);
  }
}

static level8( result)
  number *result;
{
  number hold;

  level9( result);
  if( *token == '^')
  {
    get_token();
    level9( &hold);
    arith("^", result, &hold);
  }
}

static level9( result)
  number *result;
{
  register char op;

  op = 0;
  if( (tok_type == DELIMETER) && (*token == '+' || *token == '-' ||
      *token == '!'))
  {
    op = *token;
    get_token();
  }
  level10( result);
  if( op)
    unary( op, result);
}

static level10( result)
  number *result;
{
  if( (*token == '[') && (tok_type == DELIMETER))
  {
    get_token();
    level6( result);
    if( *token != ']')
      (*error)(UNBAL_PARENT);
    get_token();
  }
  else
    primitive( result);
}

static primitive( result)
  number *result;
{

  switch (tok_type)
  {
    case VARIABLE:
                *result = find_var( token);
                return get_token();
    case NUMBER:
                *result = get_int( token);
                return get_token();
    default:
                (*error)(SYNTAX_ERROR);
  }
}

static arith( o, r, h)
  char  *o;
  number *r,
        *h;
{
  number t, ex;

  if( !strcmp( o, "-"))
    *r = *r - *h;
  else if( !strcmp( o, "+"))
    *r = *r + *h;
  else if( !strcmp( o, "*"))
    *r = *r * *h;
  else if( !strcmp( o, "/"))
    *r = (*r) / (*h);
  else if( !strcmp( o, "%"))
    *r = (int) *r % (int) *h;
  else if( !strcmp( o, "<>"))
    *r = (*r != *h);
  else if( !strcmp( o, "&"))
    *r = (int) *r & (int) *h;
  else if( !strcmp( o, "|"))
    *r = (int) *r | (int) *h;
  else if( !strcmp( o, "~"))
    *r = (int) *r ^ (int) *h;
  else if( !strcmp( o, "="))
    *r = (*r == *h);
  else if( !strcmp( o, "<"))
    *r = (*r < *h);
  else if( !strcmp( o, ">"))
    *r = (*r < *h);
  else if( !strcmp( o, "<="))
    *r = (*r <= *h);
  else if( !strcmp( o, ">="))
    *r = (*r >= *h);
  else if( !strcmp( o, "<<"))
    *r = (int) *r << (int) *h;
  else if( !strcmp( o, ">>"))
    *r = (int) *r >> (int) *h;
  else if( !strcmp( o, "^"))
  {
    ex = *r;
    if( *h == 0)
      *r = 1;
    else
    for( t= *h-1; t>0; --t)
      *r = (*r) * ex;
  }
}

static unary( o, r)
  char  o;
  number *r;
{
  switch (o)
  {
    case '-':
              *r = -(*r);
              break;
    case '!':
              *r = ! (int) (*r);
              break;
  }

}

static putback()    /* Return a token to its resting place */
{
  char *t;

  t= token;
  for(; *t; t++)
    evaexpress--;
}

static serror( error)
  int error;
{
  static char *e[]=
  {
    "Syntax Error",
    "Unbalanced Parentheses",
    "No Expression Present",
    "Variable undeclared"
  };

  printf("\n*** %s (EVALD)\n", e[error]);
  longjmp(errenv,1);
}

static get_token()
{
  register char *temp,
                *temp1,
                c;
  number        *tempval;

  tok_type= 0;
  temp= token;

  while( iswhite( *evaexpress))
    ++evaexpress;    /* Skip over white space */

  if( is_in( *evaexpress, "+-*/[]=&|<>%^~!"))
  {
    tok_type= DELIMETER;
    switch (*evaexpress)
    {
      case '-':
               if( *(evaexpress+1) == '>')
                 *temp++ = *evaexpress++;
               break;
      case '<':
               if( (c = *(evaexpress+1)) == '>' || c == '<' || c == '=' || c == '-')
                 *temp++ = *evaexpress++;
               break;
      case '>':
               if( (c = *(evaexpress+1)) == '>' || c == '=')
                 *temp++ = *evaexpress++;
               break;
    }
    /* Advance to next position */
    *temp++ = *evaexpress++;
  }
  else if( issymbol(*evaexpress))
  {
    while( !isdelim( *evaexpress))
      *temp++ = *evaexpress++;
/*
**    if( *evaexpress == '[')
**    {
**      evaexpress++;
**
**      tok_type= FUNCTION;
**      level10( tempval);
**      do_func();
*/
      tok_type= VARIABLE;
  }
  else if( isdigit( *evaexpress) || *evaexpress == '$')
  {
    while( !isdelim( *evaexpress))
      *temp++ = *evaexpress++;
    tok_type= NUMBER;
  }
  else if (evaexpress[0] == '\'' && evaexpress[2] == '\'')
  {
    *temp++ = *evaexpress++;
    *temp++ = *evaexpress++;
    evaexpress++;
    tok_type= NUMBER;
  }
  *temp= 0;
}

static number find_var( s)
  char *s;
{
  number temp;


  if( !issymbol(*s) )
    (*error)( UNBAL_PARENT);

  if (get_symbol(s, &temp) != -1)
     return temp;

  strcpy( errmess, token);
  (*error)(UNDECLD_VARI);
}

static iswhite(c)
  char c;
{
  /* Look for spaces and tabs */
  if( c == ' ' || c == '\t')
    return 1;
  return 0;
}

static isdelim(c)
  char c;
{
  if( is_in( c, " +-/*%^<>=[]~&|") || c == '\t' || c == '\n' || c == 0)
    return 1;
  return 0;
}

static is_in( ch, s)
 char ch,
      *s;
{
  while( *s)
    if( *s++ == ch)
      return 1;
  return 0;
}


static get_int(str)
char *str;
{
   int value;

   switch (str[0])
   {
      case '$':
         sscanf(str+1, "%x", &value);
         return value;

      case '\'':
         return str[1];

      case '0':
         switch (str[1])
         {
            case 'x':
            case 'X':
               sscanf(str+2, "%x", &value);
               return value;

            case 'o':
            case 'O':
               sscanf(str+2, "%o", &value);
               return value;
         }

      default:
         sscanf(str, "%d", &value);
   }

   return value;
}


static issymbol(chr)
int chr;
{
   return isalpha(chr) || chr == '@' || chr == '_';
}

